package de.hftstuttgart.unifiedticketing.systems.gitlab;

import de.hftstuttgart.unifiedticketing.core.Logging;
import de.hftstuttgart.unifiedticketing.core.TicketSystemBuilder;
import de.hftstuttgart.unifiedticketing.exceptions.AssertionException;

import java.util.logging.Level;
import java.util.logging.Logger;

public class GitlabTicketSystemBuilder extends TicketSystemBuilder<GitlabTicketSystemBuilder, GitlabTicketSystem>
{
    private static Logger logger = Logging.getLogger(GitlabTicketSystemBuilder.class.getName());

    protected String apiKey;
    protected String apiVersion;
    protected int projectId;
    protected boolean https;

    /**
     * starts builder process for a {@link GitlabTicketSystem} instance
     */
    public GitlabTicketSystemBuilder()
    {
        super();

        apiVersion = "v4";
        projectId = -1;
        https = true;

        logger.log(Level.FINEST, "Builder for Gitlab System instance started");
    }

    /**
     * defines api-key to be used for requests<br/>
     * <br/>
     * if you don't specify an api-key, all requests will be without credentials
     * @param apiKey your user specific
     * @return this ticket system builder
     */
    public GitlabTicketSystemBuilder withApiKey(String apiKey)
    {
        this.apiKey = apiKey;
        logger.log(Level.FINEST, "set api key");
        return this;
    }

    /**
     * set protocol to http (no transport security!)
     * @return this ticket system builder
     */
    public GitlabTicketSystemBuilder withHttp()
    {
        this.https = false;
        logger.log(Level.FINEST, "set http");
        return this;
    }

    /**
     * set protocol to https (enabled transport security)<br/>
     * this is the default
     * @return this ticket system builder
     */
    public GitlabTicketSystemBuilder withHttps()
    {
        this.https = true;
        logger.log(Level.FINEST, "set https");
        return this;
    }

    /**
     * set project id the instance should work in
     * @return this ticket system builder
     */
    public GitlabTicketSystemBuilder withProjectId(int projectId)
    {
        this.projectId = projectId;
        logger.log(Level.FINEST, "set project id to " + this.projectId);
        return this;
    }

    /**
     * set project id the instance should work in
     * @return this ticket system builder
     * @throws AssertionException if {@code projectId} is not parsable to as integer
     */
    public GitlabTicketSystemBuilder withProjectId(String projectId)
    {
        try
        {
            return this.withProjectId(Integer.parseInt(projectId));
        }
        catch (NumberFormatException e)
        {
            logger.log(Level.SEVERE, String.format("given String is not parsable to an integer projectId: %s", projectId));
            throw new AssertionException(e);
        }
    }

    /**
     * builds a new gitlab ticket system instance with the given information.<br/>
     * <br/>
     * The minimum builder methods needed, are {@link #withBaseUrl} and {@link #withProjectId}.
     * @return new {@link GitlabTicketSystem} instance
     * @throws AssertionException if {@link #withBaseUrl} or {@link #withProjectId} not called beforehand
     */
    @Override
    public GitlabTicketSystem build()
    {
        if (
            baseUrl == null
            || projectId == -1
        )
        {
            logger.log(Level.SEVERE, "to build a GitlabTicketSystem instance at least a base url and projectId are needed!");
            throw new AssertionException("baseUrl or projectId are missing");
        }

        GitlabTicketSystem system = new GitlabTicketSystem();

        system.baseUrl = String.format(
            "%s://%s/api/%s/projects/%d",
            (this.https) ? "https" : "http",
            this.baseUrl,
            this.apiVersion,
            this.projectId);

        system.apiKey = apiKey;

        logger.log(Level.FINEST, "created instance of " + system.getClass().getSimpleName()
            + " with baseUrl " + system.baseUrl);

        return system;
    }
}
