package de.hftstuttgart.unifiedticketing.systems.gitlab;

import com.fasterxml.jackson.databind.ObjectMapper;
import de.hftstuttgart.unifiedticketing.exceptions.AssertionException;
import de.hftstuttgart.unifiedticketing.exceptions.DeserializationException;
import de.hftstuttgart.unifiedticketing.exceptions.HttpResponseException;
import okhttp3.*;
import okio.Buffer;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;

import java.io.IOException;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

public class GitlabTicketBuilderTest
{
    public GitlabTicketBuilder instance;
    public Call call;
    public ArgumentCaptor<Request> requestCaptor;
    public Response.Builder responseBuilder;

    @BeforeEach
    public void initBeforeEach()
    {
        GitlabTicketSystem parent = mock(GitlabTicketSystem.class);
        parent.baseUrl = "https://example.org";
        instance = spy(new GitlabTicketBuilder(parent));

        OkHttpClient client = mock(OkHttpClient.class);
        call = mock(Call.class);
        requestCaptor = ArgumentCaptor.forClass(Request.class);
        responseBuilder = new Response.Builder()
            .request(new Request.Builder().url("http://test.some.tld/").build())
            .protocol(Protocol.HTTP_1_1)
            .message("some message");

        doReturn(client).when(instance).getHttpClient();
        doReturn(call).when(client).newCall(requestCaptor.capture());
    }

    @Test
    public void testSaveNoCreationWithoutMinimumRequirements()
    {
        assertThrows(AssertionException.class, () -> instance.create());
        verify(instance, never()).getHttpClient();
    }

    @Test
    public void testSaveServerError() throws IOException
    {
        doReturn(responseBuilder.code(500).build()).when(call).execute();
        instance.title("we have a new title");
        assertThrows(HttpResponseException.class, () -> instance.create());
    }

    @Test
    public void testSaveClientError() throws IOException
    {
        doReturn(responseBuilder.code(400).build()).when(call).execute();
        instance.title("we have another new title");
        assertThrows(HttpResponseException.class, () -> instance.create());
    }

    @Test
    public void testSaveNullBody() throws IOException
    {
        doReturn(responseBuilder.code(200).build()).when(call).execute();
        instance.title("we have no body this time");
        assertThrows(HttpResponseException.class, () -> instance.create());
    }

    @Test
    public void testSaveNoJsonBody() throws IOException
    {
        doReturn(
            responseBuilder
                .code(200)
                .body(ResponseBody.create("somestrangething", MediaType.get("application/json")))
                .build())
            .when(call).execute();
        instance.title("this body is no json");
        assertThrows(DeserializationException.class, () -> instance.create());
    }

    @Test
    public void testSaveSuccessfulUpdate() throws IOException
    {
        GitlabTicketResponse ticketResponse = new GitlabTicketResponse();
        ticketResponse.title = "title of ticket";
        ticketResponse.description = "description";
        ticketResponse.iid = 5;
        ticketResponse.labels = new HashSet<>(Arrays.asList("bug", "unifiedticketing"));
        GitlabTicketResponse.Assignee assignee = new GitlabTicketResponse.Assignee();
        assignee.id = 234;
        assignee.name = "Full name";
        assignee.username = "username";
        ticketResponse.assignees = new LinkedList<>(Collections.singleton(assignee));

        GitlabTicket expected = GitlabTicket.fromTicketResponse(instance.parent, ticketResponse);
        String responseJson = new ObjectMapper().writeValueAsString(ticketResponse);

        doReturn(
            responseBuilder
                .code(200)
                .body(ResponseBody.create(responseJson.getBytes(), MediaType.get("application/json")))
                .build())
            .when(call).execute();
        GitlabTicket actual = instance.title("some title").create();
        assertEquals(expected, actual);
        assertTrue(expected.deepEquals(actual));
    }

    @Test
    public void testSaveRequestJson() throws IOException
    {
        doReturn(
            responseBuilder
                .code(200)
                .body(ResponseBody.create("{}".getBytes(), MediaType.get("application/json")))
                .build())
            .when(call).execute();

        instance.title("title of ticket")
            .description("description")
            .labels(new HashSet<>(Arrays.asList("bug", "unifiedticketing")))
            .assignees(234)
            .create();

        Buffer buffer = new Buffer();
        ObjectMapper mapper = new ObjectMapper();
        requestCaptor.getValue().body().writeTo(buffer);
        String expectedJson = "{\"assignee_ids\":[234],\"description\":\"description\"," +
            "\"title\":\"title of ticket\",\"labels\":\"bug,unifiedticketing\"}";
        assertEquals(mapper.readTree(expectedJson), mapper.readTree(buffer.readUtf8()));
    }
}
